#include <QGuiApplication>
#include <QPainter>
#include <Qt3DCore/QEntity>
#include <Qt3DCore/QTransform>
#include <Qt3DRender/QCamera>
#include <Qt3DExtras/QTextureMaterial>
#include <Qt3DExtras/QSphereMesh>
#include <Qt3DExtras/QOrbitCameraController>
#include <Qt3DExtras/Qt3DWindow>
#include <Qt3DRender/QPaintedTextureImage>
#include <Qt3DExtras/QPlaneMesh>
#include <Qt3DRender/QTexture>
#include <Qt3DLogic/QFrameAction>

class TextImage : public Qt3DRender::QPaintedTextureImage
{
public:
    void paint(QPainter *painter) override;

private:
    int m_count = 0;
};

void TextImage::paint(QPainter *painter)
{
    const QSize sz = size();
    painter->setCompositionMode(QPainter::CompositionMode_Source);
    painter->fillRect(0, 0, sz.width(), sz.height(), Qt::transparent);
    painter->setCompositionMode(QPainter::CompositionMode_SourceOver);

    QPen pen(Qt::black);
    pen.setWidth(4);
    painter->setPen(pen);
    painter->drawRect(0, 0, sz.width(), sz.height());

    painter->setPen(Qt::gray);
    QFont font;
    font.setPointSize(24);
    painter->setFont(font);
    const QString s = QString::asprintf("hello world #%d", m_count);
    ++m_count;
    painter->drawText(50, 50, s);
}

Qt3DCore::QEntity *createScene()
{
    const QSize sz(320, 200);

    Qt3DCore::QEntity *rootEntity = new Qt3DCore::QEntity;

    auto texture = new Qt3DRender::QTexture2D(rootEntity);
    auto textureImage = new TextImage;
    textureImage->setSize(sz);
    texture->addTextureImage(textureImage);

    Qt3DCore::QEntity *entity = new Qt3DCore::QEntity(rootEntity);

    auto mesh = new Qt3DExtras::QPlaneMesh;
    mesh->setWidth(16);
    mesh->setHeight(16);
    entity->addComponent(mesh);

    auto material = new Qt3DExtras::QTextureMaterial(rootEntity);
    material->setTexture(texture);
    material->setAlphaBlendingEnabled(true);
    entity->addComponent(material);

    auto transform = new Qt3DCore::QTransform;
    transform->setRotationX(90);
    entity->addComponent(transform);

    auto frameAction = new Qt3DLogic::QFrameAction;
    QObject::connect(frameAction, &Qt3DLogic::QFrameAction::triggered, frameAction, [textureImage] {
        textureImage->update();
    });
    rootEntity->addComponent(frameAction);

    return rootEntity;
}

int main(int argc, char* argv[])
{
    QGuiApplication app(argc, argv);
    Qt3DExtras::Qt3DWindow view;

    Qt3DCore::QEntity *scene = createScene();

    Qt3DRender::QCamera *camera = view.camera();
    camera->lens()->setPerspectiveProjection(45.0f, 16.0f/9.0f, 0.1f, 1000.0f);
    camera->setPosition(QVector3D(0, 0, 40.0f));
    camera->setViewCenter(QVector3D(0, 0, 0));

    Qt3DExtras::QOrbitCameraController *camController = new Qt3DExtras::QOrbitCameraController(scene);
    camController->setLinearSpeed( 50.0f );
    camController->setLookSpeed( 180.0f );
    camController->setCamera(camera);

    view.setRootEntity(scene);
    view.show();

    return app.exec();
}
